==> kahakai-0.6.1.tar.bz2: kahakai-0.6.1/src/ImageControl.cc <==
/**
 * @file   ImageControl.cc
 * @date   Sat Oct  4 23:19:22 2003
 * 
 * @brief Implementation of ImageControl class which handles rendering
 *        drawing operations into pixmap.
 * 
 * 
 */

#include "ImageControl.hh"
#include "Texture.hh"
#include "Screen.hh"

#include <map>
#include <utility>
#include <vector>

using std::make_pair;
using std::map;
using std::max;
using std::pair;
using std::vector;

static const unsigned int CACHE_MAX = 15;

ImageControl::ImageControl(KaScreen* _kascreen, Display *_display, int screen,
                           Visual *_visual, 
                           Colormap _colormap) {

    display = _display;
    kascreen = _kascreen;

    Colormap colormap      = DefaultColormap(kascreen->display, 
                                             kascreen->screen_number);
    Visual*  visual        = DefaultVisual(kascreen->display, 
                                           kascreen->screen_number);

    ctx = new ImContext(display, visual, colormap);
    root_atom = XInternAtom(display, "_XROOTPMAP_ID", 1);
    this->screen = screen;

    //rootPixmap = kascreen->xrootpmap_id;
    //    if (!rootPixmap) rootPixmap = None;
    getRootImage();
}

ImageControl::~ImageControl() {
    delete ctx;
}

void ImageControl::getRootImage() {
    int real_format;
    Atom real_type;
    CARD32 *data;   

    unsigned long items_read, items_left;
    if (XGetWindowProperty(display, RootWindow(display, screen), root_atom,
                           0L, 1L, 
                           false, XA_PIXMAP, &real_type,
                           &real_format, &items_read, &items_left, 
                           (unsigned char **) &data) == Success && 
                           items_read) {
        rootPixmap = (Pixmap)(*data);
        free(data);
    }
    else
        rootPixmap = None;
}

Pixmap ImageControl::render(Drawable parent, const DrawOps& ops, 
                            int width, int height) {
    ImImage img(width, height);
    Pixmap pixmap, mask;

    for(DrawOps::const_iterator it = ops.begin(); it != ops.end(); ++it) {
        (*it)->draw(img);
    }

    ctx->push();
    img.render_pixmaps(parent, &pixmap, &mask);
    ctx->pop();

    return pixmap;
}

Pixmap ImageControl::render(Drawable parent, const DrawOps &ops, 
                            int x, int y, int width, int height) {
    if (rootPixmap == None)
        return render(parent, ops, width, height);

    return render(parent, TexturePtr(new Texture(ops)), x, y, width, height);
}

shared_ptr<ImImage> ImageControl::getTextureImage(TexturePtr texture,
                                                  int width, int height) {

    if (!texture) return shared_ptr<ImImage>();

    // Try to find it in the cache
    pair<int, int> key(width, height);

    shared_ptr<ImImage> texture_img;

    if (cache[texture].count(key) > 0 && cache[texture][key]) {
        texture_img = cache[texture][key];
        cacheCount[texture][key]++;
    } else {
        texture_img = texture->render(width, height);
        cache[texture][key] = texture_img;
    }

    cleanup();

    if (cache[texture].size() > CACHE_MAX) {
        timeout();
    }

    return texture_img;
}

shared_ptr<ImImage> ImageControl::renderToImage(TexturePtr texture,
                                              int x, int y, int width, 
                                              int height) {
    if (!texture) return shared_ptr<ImImage>();

    shared_ptr<ImImage> texture_img = getTextureImage(texture, width, height);
    shared_ptr<ImImage> result;

    if (rootPixmap != None) {
        ctx->push();
        if (x < 0 || y < 0) {
            if ((-x) > width || (-y) > height) return shared_ptr<ImImage>();

            result = shared_ptr<ImImage>(new ImImage(width, height));
            ImImage background(rootPixmap, 0, 0, 0, 
                               width - (-x), height - (-y), true);

            imlib_context_set_blend(0);
            result->blend(background, false, -x, -y);
            imlib_context_set_blend(1);
        } else {
            result = shared_ptr<ImImage>(new ImImage
                                         (rootPixmap, 0, x, y, width, height, 
                                          false));
        }

        result->blend(*texture_img, false, 0, 0);
        ctx->pop();        
    } else {
        result = texture_img;
    }

    return result;
}

Pixmap ImageControl::renderImage(Drawable parent, shared_ptr<ImImage> image) {
    Pixmap pixmap, mask;
    ctx->push();
    image->render_pixmaps(parent, &pixmap, &mask);
    ctx->pop();

    return pixmap;
}

void ImageControl::renderImageOnDrawable(Drawable drawable, 
                                         shared_ptr<ImImage> image,
                                         int x, int y) {
    ctx->push();
    image->render_on_drawable(drawable, x, y);
    ctx->pop();
}
                                 

Pixmap ImageControl::render(Drawable parent, TexturePtr texture, 
                            int x, int y, int width, int height) {
    // If no texture, then it's fully transparent.
    if (!texture) return ParentRelative;

    ctx->push();
    // Try to find it in the cache
    shared_ptr<ImImage> texture_img = getTextureImage(texture, width, height);

    Pixmap pixmap, mask;

    if (rootPixmap != None) {
        if (x < 0 || y < 0) {
            if ((-x) > width || (-y) > height) return ParentRelative;
            
            // If part of the window is off screen, we need to get only
            // relevant part of background pixmap and offset it in
            // opposite direction
            ImImage temp(width, height);
            ImImage background(rootPixmap, 0, 0, 0, 
                               width - (-x), height - (-y), true);

            // Don't need blending as we're just copying the image to temp
            imlib_context_set_blend(0);
            temp.blend(background,   false, -x, -y);
            imlib_context_set_blend(1);

            temp.blend(*texture_img, false, 0, 0);
            temp.render_pixmaps(parent, &pixmap, &mask);
        } else {
            ImImage background(rootPixmap, 0, max(x, 0), max(y, 0), 
                               width, height, true);
            background.blend(*texture_img, false, 0, 0);
            background.render_pixmaps(parent, &pixmap, &mask);
        }
    } else {
        texture_img->render_pixmaps(parent, &pixmap, &mask);
    }

    ctx->pop();
    return pixmap;
}

void ImageControl::render_on_drawable(Drawable drawable,
                                      TexturePtr texture,
                                      int x, int y, 
                                      unsigned int width, 
                                      unsigned int height) {
    if (!texture) return;

    shared_ptr<ImImage> texture_img = getTextureImage(texture, width, height);
    texture_img->render_on_drawable(drawable, x, y);
}

void ImageControl::cleanup() {
    Cache::iterator it;
    for (it = cache.begin(); it != cache.end(); ++it) {
        if (it->first.expired() || it->second.empty())
            cache.erase(it);
    }
}

void ImageControl::timeout() {
    Cache::iterator it;
    for (it = cache.begin(); it != cache.end(); ++it) {
        map<pair<int, int>, shared_ptr<ImImage> >::iterator it2;

        if (it->second.size() <= CACHE_MAX) continue;

        // Sort by use count.

        // We erase quarter of the entries to avoid constant timeout()
        // calls.
        vector<pair<int, pair<int, int> > > entries;
        unsigned int n = it->second.size();
        for (it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
            entries.push_back(make_pair(
                                        cacheCount[it->first][it2->first], 
                                        it2->first));
            n--;
            if (n < CACHE_MAX) break;
        }
        sort(entries.begin(), entries.end());

        for (unsigned int i=0;i<entries.size();i++) {
            it->second.erase(it->second.find(entries[i].second));
            cacheCount[it->first].erase(cacheCount[it->first].find(entries[i].second));
        }
    }
}

void ImageControl::clearCache() {
    cache.clear();

    getRootImage();
}

==> kahakai-0.6.2.tar.bz2: kahakai-0.6.2/src/ImageControl.cc <==
/**
 * @file   ImageControl.cc
 * @date   Sat Oct  4 23:19:22 2003
 * 
 * @brief Implementation of ImageControl class which handles rendering
 *        drawing operations into pixmap.
 * 
 * 
 */

#include "ImageControl.hh"
#include "Texture.hh"
#include "Screen.hh"

#include <map>
#include <utility>
#include <vector>

using std::make_pair;
using std::map;
using std::max;
using std::pair;
using std::vector;

static const unsigned int CACHE_MAX = 15;

ImageControl::ImageControl(KaScreen* _kascreen, Display *_display, int screen,
                           Visual *_visual, 
                           Colormap _colormap) {

    display = _display;
    kascreen = _kascreen;

    Colormap colormap      = DefaultColormap(kascreen->display, 
                                             kascreen->screen_number);
    Visual*  visual        = DefaultVisual(kascreen->display, 
                                           kascreen->screen_number);

    ctx = new ImContext(display, visual, colormap);
    root_atom = XInternAtom(display, "_XROOTPMAP_ID", 1);
    this->screen = screen;

    //rootPixmap = kascreen->xrootpmap_id;
    //    if (!rootPixmap) rootPixmap = None;
    getRootImage();
}

ImageControl::~ImageControl() {
    delete ctx;
}

void ImageControl::getRootImage() {
    int real_format;
    Atom real_type;
    CARD32 *data;   

    unsigned long items_read, items_left;
    if (XGetWindowProperty(display, RootWindow(display, screen), root_atom,
                           0L, 1L, 
                           false, XA_PIXMAP, &real_type,
                           &real_format, &items_read, &items_left, 
                           (unsigned char **) &data) == Success && 
                           items_read) {
        rootPixmap = (Pixmap)(*data);
        free(data);
    }
    else
        rootPixmap = None;
}

Pixmap ImageControl::render(Drawable parent, const DrawOps& ops, 
                            int width, int height) {
    ImImage img(width, height);
    Pixmap pixmap, mask;

    for(DrawOps::const_iterator it = ops.begin(); it != ops.end(); ++it) {
        (*it)->draw(img);
    }

    ctx->push();
    img.render_pixmaps(parent, &pixmap, &mask);
    ctx->pop();

    return pixmap;
}

Pixmap ImageControl::render(Drawable parent, const DrawOps &ops, 
                            int x, int y, int width, int height) {
    if (rootPixmap == None)
        return render(parent, ops, width, height);

    return render(parent, TexturePtr(new Texture(ops)), x, y, width, height);
}

shared_ptr<ImImage> ImageControl::getTextureImage(TexturePtr texture,
                                                  int width, int height) {

    if (!texture) return shared_ptr<ImImage>();

    // Try to find it in the cache
    pair<int, int> key(width, height);

    shared_ptr<ImImage> texture_img;

    if (cache[texture].count(key) > 0 && cache[texture][key]) {
        texture_img = cache[texture][key];
        cacheCount[texture][key]++;
    } else {
        texture_img = texture->render(width, height);
        cache[texture][key] = texture_img;
    }

    cleanup();

    if (cache[texture].size() > CACHE_MAX) {
        timeout();
    }

    return texture_img;
}

shared_ptr<ImImage> ImageControl::renderToImage(TexturePtr texture,
                                              int x, int y, int width, 
                                              int height) {
    if (!texture) return shared_ptr<ImImage>();

    shared_ptr<ImImage> texture_img = getTextureImage(texture, width, height);
    shared_ptr<ImImage> result;

    if (rootPixmap != None) {
        ctx->push();
        if (x < 0 || y < 0) {
            if ((-x) > width || (-y) > height) return shared_ptr<ImImage>();

            result = shared_ptr<ImImage>(new ImImage(width, height));
            ImImage background(rootPixmap, 0, 0, 0, 
                               width - (-x), height - (-y), true);

            imlib_context_set_blend(0);
            result->blend(background, false, -x, -y);
            imlib_context_set_blend(1);
        } else {
            result = shared_ptr<ImImage>(new ImImage
                                         (rootPixmap, 0, x, y, width, height, 
                                          false));
        }

        result->blend(*texture_img, false, 0, 0);
        ctx->pop();        
    } else {
        result = texture_img;
    }

    return result;
}

Pixmap ImageControl::renderImage(Drawable parent, shared_ptr<ImImage> image) {
    Pixmap pixmap, mask;
    ctx->push();
    image->render_pixmaps(parent, &pixmap, &mask);
    ctx->pop();

    return pixmap;
}

void ImageControl::renderImageOnDrawable(Drawable drawable, 
                                         shared_ptr<ImImage> image,
                                         int x, int y) {
    ctx->push();
    image->render_on_drawable(drawable, x, y);
    ctx->pop();
}
                                 

Pixmap ImageControl::render(Drawable parent, TexturePtr texture, 
                            int x, int y, int width, int height) {
    // If no texture, then it's fully transparent.
    if (!texture) return ParentRelative;

    ctx->push();
    // Try to find it in the cache
    shared_ptr<ImImage> texture_img = getTextureImage(texture, width, height);

    Pixmap pixmap, mask;

    if (rootPixmap != None) {
        if (x < 0 || y < 0) {
            if ((-x) > width || (-y) > height) {
            	ctx->pop();
            	return ParentRelative;
            }
            
            // If part of the window is off screen, we need to get only
            // relevant part of background pixmap and offset it in
            // opposite direction
            ImImage temp(width, height);
            ImImage background(rootPixmap, 0, 0, 0, 
                               width - (-x), height - (-y), true);

            // Don't need blending as we're just copying the image to temp
            imlib_context_set_blend(0);
            temp.blend(background,   false, -x, -y);
            imlib_context_set_blend(1);

            temp.blend(*texture_img, false, 0, 0);
            temp.render_pixmaps(parent, &pixmap, &mask);
        } else {
            ImImage background(rootPixmap, 0, max(x, 0), max(y, 0), 
                               width, height, true);
            background.blend(*texture_img, false, 0, 0);
            background.render_pixmaps(parent, &pixmap, &mask);
        }
    } else {
        texture_img->render_pixmaps(parent, &pixmap, &mask);
    }

    ctx->pop();
    return pixmap;
}

void ImageControl::render_on_drawable(Drawable drawable,
                                      TexturePtr texture,
                                      int x, int y, 
                                      unsigned int width, 
                                      unsigned int height) {
    if (!texture) return;

    shared_ptr<ImImage> texture_img = getTextureImage(texture, width, height);
    texture_img->render_on_drawable(drawable, x, y);
}

void ImageControl::cleanup() {
    Cache::iterator it;
    for (it = cache.begin(); it != cache.end(); ++it) {
        if (it->first.expired() || it->second.empty())
            cache.erase(it);
    }
}

void ImageControl::timeout() {
    Cache::iterator it;
    for (it = cache.begin(); it != cache.end(); ++it) {
        map<pair<int, int>, shared_ptr<ImImage> >::iterator it2;

        if (it->second.size() <= CACHE_MAX) continue;

        // Sort by use count.

        // We erase quarter of the entries to avoid constant timeout()
        // calls.
        vector<pair<int, pair<int, int> > > entries;
        unsigned int n = it->second.size();
        for (it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
            entries.push_back(make_pair(
                                        cacheCount[it->first][it2->first], 
                                        it2->first));
            n--;
            if (n < CACHE_MAX) break;
        }
        sort(entries.begin(), entries.end());

        for (unsigned int i=0;i<entries.size();i++) {
            it->second.erase(it->second.find(entries[i].second));
            cacheCount[it->first].erase(cacheCount[it->first].find(entries[i].second));
        }
    }
}

void ImageControl::clearCache() {
    cache.clear();

    getRootImage();
}
